#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# _____________________________________________________________________
#
# Shortcuts 
#   * imported by import order.
#   * should be imported as.
# 		from dev0s.shortcuts import *
# 
#

# default pip imports.
import os, sys

# import.
import dev0s

# utils.
utils = dev0s.utils

# exceptions.
exceptions = dev0s.exceptions

# color.
color = dev0s.color
symbol = dev0s.symbol

# console.
Loader = dev0s.console.Loader
ProgressLoader = dev0s.console.ProgressLoader

# files & formats.
Files = dev0s.files.Files
Formats = dev0s.files.Formats
gfp = dev0s.files.gfp
gdate = dev0s.files.gdate

FilePath = dev0s.files.FilePath 
String = dev0s.files.String 
Boolean = dev0s.files.Boolean 
Integer = dev0s.files.Integer 
Date = dev0s.files.Date

File = dev0s.files.File
Directory = dev0s.files.Directory
Zip = dev0s.files.Zip
Image = dev0s.files.Image
Bytes = dev0s.files.Bytes
Dictionary = dev0s.files.Dictionary
Array = dev0s.files.Array

Generate = dev0s.files.Generate
Interval = dev0s.files.Interval
Speed = dev0s.files.Speed
speed = dev0s.files.Speed()


# response.
ResponseObject = dev0s.response.ResponseObject

# objects.
Traceback = dev0s.objects.Traceback
Object = dev0s.objects.Object
Thread = dev0s.objects.Thread
ThreadManager = dev0s.objects.ThreadManager

# class database.
Database = dev0s.database.Database

# requests.
RestAPI = dev0s.requests.RestAPI

# code.
OutputObject = dev0s.code.OutputObject
Spawn = dev0s.code.Spawn
Version = dev0s.code.Version 
Script = dev0s.code.Script
Python = dev0s.code.Python

#
